"""Engine/Runner provider."""

from __future__ import annotations

import enum
from typing import Optional, Union, List, Dict, AsyncGenerator, Mapping

from anyio import move_on_after

from opentrons.types import NozzleMapInterface
from opentrons_shared_data.labware.types import LabwareUri
from opentrons_shared_data.labware.labware_definition import LabwareDefinition
from opentrons_shared_data.errors import GeneralError
from opentrons_shared_data.robot.types import RobotType

from . import protocol_runner, RunResult, JsonRunner, PythonAndLegacyRunner
from ..hardware_control import HardwareControlAPI
from ..hardware_control.modules import AbstractModule as HardwareModuleAPI
from ..protocol_engine import (
    ProtocolEngine,
    CommandCreate,
    Command,
    StateSummary,
    CommandPointer,
    CommandSlice,
    CommandErrorSlice,
    DeckType,
    ErrorOccurrence,
)
from ..protocol_engine.errors import RunStoppedError
from ..protocol_engine.types import (
    PostRunHardwareState,
    EngineStatus,
    LabwareOffsetCreate,
    LegacyLabwareOffsetCreate,
    LabwareOffset,
    DeckConfigurationType,
    RunTimeParameter,
    PrimitiveRunTimeParamValuesType,
    CSVRuntimeParamPaths,
    CommandAnnotation,
)
from ..protocol_engine.error_recovery_policy import ErrorRecoveryPolicy

from ..protocol_reader import JsonProtocolConfig, PythonProtocolConfig, ProtocolSource
from ..protocols.parse import PythonParseMode


class NoProtocolRunAvailable(RuntimeError):
    """An error raised if there is no protocol run available."""


class UnknownProtocolParseMode(RuntimeError):
    """An error raised if given an unknown protocol parse mode."""


class RunNotFound(GeneralError):
    """An error raised if there is no run associated."""


class ParseMode(enum.Enum):
    """Configure optional rules for when `opentrons.protocols.parse.parse()` parses protocols."""

    NORMAL = enum.auto()
    ALLOW_LEGACY_METADATA_AND_REQUIREMENTS = enum.auto()


class RunOrchestrator:
    """Provider for runners and associated protocol engine.

    Build runners, manage command execution, run state and in-memory protocol engine associated to the runners.
    """

    _protocol_runner: Optional[
        Union[protocol_runner.JsonRunner, protocol_runner.PythonAndLegacyRunner, None]
    ]
    _setup_runner: protocol_runner.LiveRunner
    _fixit_runner: protocol_runner.LiveRunner
    _protocol_live_runner: protocol_runner.LiveRunner
    _hardware_api: HardwareControlAPI
    _protocol_engine: ProtocolEngine

    def __init__(
        self,
        protocol_engine: ProtocolEngine,
        # todo(mm, 2024-07-05): This hardware_api param looks unused?
        hardware_api: HardwareControlAPI,
        fixit_runner: protocol_runner.LiveRunner,
        setup_runner: protocol_runner.LiveRunner,
        protocol_live_runner: protocol_runner.LiveRunner,
        json_or_python_protocol_runner: Optional[
            Union[protocol_runner.PythonAndLegacyRunner, protocol_runner.JsonRunner]
        ] = None,
        run_id: Optional[str] = None,
    ) -> None:
        """Initialize a run orchestrator interface.

        Arguments:
            protocol_engine: Protocol engine instance.
            hardware_api: Hardware control API instance.
            fixit_runner: LiveRunner for fixit commands.
            setup_runner: LiveRunner for setup commands.
            protocol_live_runner: LiveRunner for protocol commands.
            json_or_python_protocol_runner: JsonRunner/PythonAndLegacyRunner for protocol commands.
            run_id: run id if any, associated to the runner/engine.
        """
        self._run_id = run_id
        self._protocol_engine = protocol_engine
        self._protocol_runner = json_or_python_protocol_runner
        self._setup_runner = setup_runner
        self._fixit_runner = fixit_runner
        self._protocol_live_runner = protocol_live_runner
        self._fixit_runner.prepare()
        self._setup_runner.prepare()
        self._protocol_engine.set_and_start_queue_worker(self.command_generator)

    @property
    def run_id(self) -> str:
        """Get the "current" persisted run_id."""
        if not self._run_id:
            raise RunNotFound()
        return self._run_id

    @classmethod
    def build_orchestrator(
        cls,
        hardware_api: HardwareControlAPI,
        protocol_engine: ProtocolEngine,
        protocol_config: Optional[
            Union[JsonProtocolConfig, PythonProtocolConfig]
        ] = None,
        post_run_hardware_state: PostRunHardwareState = PostRunHardwareState.HOME_AND_STAY_ENGAGED,
        drop_tips_after_run: bool = True,
        run_id: Optional[str] = None,
    ) -> "RunOrchestrator":
        """Build a RunOrchestrator provider."""
        setup_runner = protocol_runner.LiveRunner(
            protocol_engine=protocol_engine,
            hardware_api=hardware_api,
        )
        fixit_runner = protocol_runner.LiveRunner(
            protocol_engine=protocol_engine,
            hardware_api=hardware_api,
        )
        protocol_live_runner = protocol_runner.LiveRunner(
            protocol_engine=protocol_engine,
            hardware_api=hardware_api,
        )
        json_or_python_runner = None
        if protocol_config:
            json_or_python_runner = protocol_runner.create_protocol_runner(
                protocol_config=protocol_config,
                protocol_engine=protocol_engine,
                hardware_api=hardware_api,
                post_run_hardware_state=post_run_hardware_state,
                drop_tips_after_run=drop_tips_after_run,
            )
        return cls(
            run_id=run_id,
            json_or_python_protocol_runner=json_or_python_runner,
            setup_runner=setup_runner,
            fixit_runner=fixit_runner,
            hardware_api=hardware_api,
            protocol_engine=protocol_engine,
            protocol_live_runner=protocol_live_runner,
        )

    def play(self, deck_configuration: Optional[DeckConfigurationType] = None) -> None:
        """Start or resume the run."""
        # todo(mm, 2024-07-09): The deck configuration is set at the same time here for
        # historical reasons. It's unsafe to change the deck configuration mid-run
        # and we're relying on the caller to not do that.
        self._protocol_engine.set_deck_configuration(deck_configuration)
        self._protocol_engine.play()

    async def run(
        self,
        deck_configuration: DeckConfigurationType,
        protocol_source: Optional[ProtocolSource] = None,
        run_time_param_values: Optional[PrimitiveRunTimeParamValuesType] = None,
    ) -> RunResult:
        """Start the run."""
        if self._protocol_runner:
            return await self._protocol_runner.run(
                deck_configuration=deck_configuration,
                protocol_source=protocol_source,
                run_time_param_values=run_time_param_values,
            )
        elif self._protocol_live_runner:
            return await self._protocol_live_runner.run(
                deck_configuration=deck_configuration
            )
        else:
            return await self._setup_runner.run(deck_configuration=deck_configuration)

    def pause(self) -> None:
        """Pause the run."""
        self._protocol_engine.request_pause()

    async def stop(self) -> None:
        """Stop the run."""
        if self.run_has_started():
            await self._protocol_engine.request_stop()
        else:
            await self.finish(
                drop_tips_after_run=False,
                set_run_status=False,
                post_run_hardware_state=PostRunHardwareState.STAY_ENGAGED_IN_PLACE,
            )

    def resume_from_recovery(self, reconcile_false_positive: bool) -> None:
        """Resume the run from recovery."""
        self._protocol_engine.resume_from_recovery(reconcile_false_positive)

    async def finish(
        self,
        error: Optional[Exception] = None,
        drop_tips_after_run: bool = True,
        set_run_status: bool = True,
        post_run_hardware_state: PostRunHardwareState = PostRunHardwareState.HOME_AND_STAY_ENGAGED,
    ) -> None:
        """Finish the run."""
        await self._protocol_engine.finish(
            error=error,
            drop_tips_after_run=drop_tips_after_run,
            set_run_status=set_run_status,
            post_run_hardware_state=post_run_hardware_state,
        )

    def get_state_summary(self) -> StateSummary:
        """Get protocol run data."""
        return self._protocol_engine.state_view.get_summary()

    def get_loaded_labware_definitions(self) -> List[LabwareDefinition]:
        """Get loaded labware definitions."""
        return self._protocol_engine.state_view.labware.get_loaded_labware_definitions()

    def get_run_time_parameters(self) -> List[RunTimeParameter]:
        """Get the list of run time parameters defined in the protocol, if any.

        This returns a list of all run time parameters with their validated definitions
        and client-requested values. Will always be empty before loading the runner.

        If there was an error during RTP definition validation, then this list will
        contain the parameter definitions that were validated before the error occurred.
        These parameters' values will be default values.

        If all definitions validated successfully but an error occurred while
        setting the RTP values with those sent by the client, then only the parameters
        whose values were successfully set will have the client-requested values while
        the others will contain the default values.
        """
        return (
            []
            if self._protocol_runner is None
            else self._protocol_runner.run_time_parameters
        )

    def get_command_annotations(self) -> List[CommandAnnotation]:
        """Get the list of command annotations defined in the protocol, if any."""
        return (
            []
            if self._protocol_runner is None
            else self._protocol_runner.command_annotations
        )

    def get_current_command(self) -> Optional[CommandPointer]:
        """Get the "current" command, if any."""
        return self._protocol_engine.state_view.commands.get_current()

    def get_most_recently_finalized_command(self) -> Optional[CommandPointer]:
        """Get the most recently finalized command, if any."""
        most_recently_finalized_command = (
            self._protocol_engine.state_view.commands.get_most_recently_finalized_command()
        )
        return (
            CommandPointer(
                command_id=most_recently_finalized_command.command.id,
                command_key=most_recently_finalized_command.command.key,
                created_at=most_recently_finalized_command.command.createdAt,
                index=most_recently_finalized_command.index,
            )
            if most_recently_finalized_command
            else None
        )

    def get_command_slice(
        self, cursor: Optional[int], length: int, include_fixit_commands: bool
    ) -> CommandSlice:
        """Get a slice of run commands.

        Args:
            cursor: Requested index of first command in the returned slice.
            length: Length of slice to return.
            include_fixit_commands: Get all command intents.
        """
        return self._protocol_engine.state_view.commands.get_slice(
            cursor=cursor, length=length, include_fixit_commands=include_fixit_commands
        )

    def get_command_error_slice(
        self,
        cursor: int,
        length: int,
    ) -> CommandErrorSlice:
        """Get a slice of run commands errors.

        Args:
            cursor: Requested index of first error in the returned slice.
                If the cursor is omitted, a cursor will be selected automatically
                based on the last error occurence.
            length: Length of slice to return.
        """
        return self._protocol_engine.state_view.commands.get_errors_slice(
            cursor=cursor, length=length
        )

    def get_command_recovery_target(self) -> Optional[CommandPointer]:
        """Get the current error recovery target."""
        return self._protocol_engine.state_view.commands.get_recovery_target()

    def get_command(self, command_id: str) -> Command:
        """Get a run's command by ID."""
        return self._protocol_engine.state_view.commands.get(command_id=command_id)

    def get_all_commands(self) -> List[Command]:
        """Get all run commands."""
        return self._protocol_engine.state_view.commands.get_all()

    def get_command_errors(self) -> List[ErrorOccurrence]:
        """Get all run command errors."""
        return self._protocol_engine.state_view.commands.get_all_errors()

    def get_run_status(self) -> EngineStatus:
        """Get the current execution status of the engine."""
        return self._protocol_engine.state_view.commands.get_status()

    def get_is_run_terminal(self) -> bool:
        """Get whether engine is in a terminal state."""
        return self._protocol_engine.state_view.commands.get_is_terminal()

    def run_has_started(self) -> bool:
        """Get whether the run has started."""
        return self._protocol_engine.state_view.commands.has_been_played()

    def run_has_stopped(self) -> bool:
        """Get whether the run has stopped."""
        return self._protocol_engine.state_view.commands.get_is_stopped()

    def add_labware_offset(
        self, request: LabwareOffsetCreate | LegacyLabwareOffsetCreate
    ) -> LabwareOffset:
        """Add a new labware offset to state."""
        return self._protocol_engine.add_labware_offset(request)

    def add_labware_definition(self, definition: LabwareDefinition) -> LabwareUri:
        """Add a new labware definition to state."""
        return self._protocol_engine.add_labware_definition(definition)

    async def add_command_and_wait_for_interval(
        self,
        command: CommandCreate,
        wait_until_complete: bool = False,
        timeout: Optional[int] = None,
        failed_command_id: Optional[str] = None,
    ) -> Command:
        """Add a new command to execute and wait for it to complete if needed."""
        added_command = self._protocol_engine.add_command(
            request=command, failed_command_id=failed_command_id
        )
        if wait_until_complete:
            timeout_sec = None if timeout is None else timeout / 1000.0
            with move_on_after(timeout_sec):
                await self._protocol_engine.wait_for_command(added_command.id)
        return added_command

    def estop(self) -> None:
        """Handle an E-stop event from the hardware API."""
        return self._protocol_engine.estop()

    async def use_attached_modules(
        self, modules_by_id: Dict[str, HardwareModuleAPI]
    ) -> None:
        """Load attached modules directly into state, without locations."""
        await self._protocol_engine.use_attached_modules(modules_by_id=modules_by_id)

    def get_protocol_runner(self) -> Optional[Union[JsonRunner, PythonAndLegacyRunner]]:
        """Get run's protocol runner if any, if not return None."""
        return self._protocol_runner

    async def load(
        self,
        protocol_source: ProtocolSource,
        run_time_param_values: Optional[PrimitiveRunTimeParamValuesType],
        run_time_param_paths: Optional[CSVRuntimeParamPaths],
        parse_mode: ParseMode,
    ) -> None:
        """Load a json/python protocol."""
        assert self._protocol_runner is not None
        if isinstance(self._protocol_runner, JsonRunner):
            await self._protocol_runner.load(protocol_source=protocol_source)
        elif isinstance(self._protocol_runner, PythonAndLegacyRunner):
            python_parse_mode = self._map_parse_mode_to_python_parse_mode(parse_mode)
            await self._protocol_runner.load(
                protocol_source=protocol_source,
                # Conservatively assume that we're re-running a protocol that
                # was uploaded before we added stricter validation, and that
                # doesn't conform to the new rules.
                python_parse_mode=python_parse_mode,
                run_time_param_values=run_time_param_values,
                run_time_param_paths=run_time_param_paths,
            )

    def get_is_okay_to_clear(self) -> bool:
        """Get whether the engine is stopped or sitting idly, so it could be removed."""
        return self._protocol_engine.state_view.commands.get_is_okay_to_clear()

    def prepare(self) -> None:
        """Prepare live runner for a run."""
        self._protocol_live_runner.prepare()

    def get_robot_type(self) -> RobotType:
        """Get engine robot type."""
        return self._protocol_engine.state_view.config.robot_type

    def get_deck_type(self) -> DeckType:
        """Get engine deck type."""
        return self._protocol_engine.state_view.config.deck_type

    def get_nozzle_maps(self) -> Mapping[str, NozzleMapInterface]:
        """Get current nozzle maps keyed by pipette id."""
        return self._protocol_engine.state_view.pipettes.get_nozzle_configurations()

    def get_tip_attached(self) -> Dict[str, bool]:
        """Get current tip state keyed by pipette id."""

        def has_tip_attached(pipette_id: str) -> bool:
            return (
                self._protocol_engine.state_view.pipettes.get_attached_tip(pipette_id)
                is not None
            )

        pipette_ids = (
            pipette.id
            for pipette in self._protocol_engine.state_view.pipettes.get_all()
        )
        return {pipette_id: has_tip_attached(pipette_id) for pipette_id in pipette_ids}

    def set_error_recovery_policy(self, policy: ErrorRecoveryPolicy) -> None:
        """Create error recovery policy for the run."""
        self._protocol_engine.set_error_recovery_policy(policy)

    async def command_generator(self) -> AsyncGenerator[str, None]:
        """Yield next command to execute."""
        while True:
            try:
                # TODO(tz, 6-26-2024): avoid using private accessor in a follow up pr.
                command_id = await self._protocol_engine._state_store.wait_for(
                    condition=self._protocol_engine.state_view.commands.get_next_to_execute
                )
                # Assert for type hinting. This is valid because the wait_for() above
                # only returns when the value is truthy.
                assert command_id is not None
                yield command_id
            except RunStoppedError:
                # There are no more commands that we should execute, either because the run has
                # completed on its own, or because a client requested it to stop.
                break

    @staticmethod
    def _map_parse_mode_to_python_parse_mode(parse_mode: ParseMode) -> PythonParseMode:
        if parse_mode == ParseMode.NORMAL:
            return PythonParseMode.NORMAL
        elif parse_mode == ParseMode.ALLOW_LEGACY_METADATA_AND_REQUIREMENTS:
            return PythonParseMode.ALLOW_LEGACY_METADATA_AND_REQUIREMENTS
        else:
            raise UnknownProtocolParseMode()



from serial import Serial  # type: ignore[import-untyped]
import asyncio
import subprocess
from typing import Any

_READ_ALL = "readall"
_READ_LINE = "read"
_DONE = "done"
_MOVE_SEAL = "ms"
_MOVE_LID = "ml"

gcode_shortcuts = {
    "status": "M119",
    _MOVE_SEAL: "M241.D",  # move seal motor
    _MOVE_LID: "M240.D",  # move lid stepper motor
    "ol": "M126",  # open lid
    "cl": "M127",  # close lid
    "sw": "M901.D",  # status of all switches
    "lt": "M141.D",  # get lid temperature
    "pt": "M105.D",  # get plate temperature
}


async def message_read(dev: Serial) -> Any:
    response = dev.readline().decode()
    while not response:
        await asyncio.sleep(1)
        response = dev.readline().decode()
    return response


async def message_return(dev: Serial) -> Any:
    try:
        response = await asyncio.wait_for(message_read(dev), timeout=20)
        return response
    except asyncio.exceptions.TimeoutError:
        print("response timed out.")
        return ""


async def handle_gcode_shortcut(dev: Serial, command: str) -> None:
    # handle debugging commands that require followup
    if command == _MOVE_SEAL:
        distance = input("enter distance in steps => ")
        dev.write(
            f"{gcode_shortcuts[command]} {distance}\n".encode()
        )  # (+) -> retract, (-) -> engage
        print(await message_return(dev))
    elif command == _MOVE_LID:
        distance = input(
            "enter angular distance in degrees => "
        )  # (+) -> open, (-) -> close
        dev.write(f"{gcode_shortcuts[command]} {distance}\n".encode())
        print(await message_return(dev))
    # everything else
    else:
        dev.write(f"{gcode_shortcuts[command]}\n".encode())
        print(await message_return(dev))


async def comms_loop(dev: Serial) -> bool:
    _exit = False
    command = input("\n>>> ")
    if command == _READ_ALL:
        print(dev.readlines())
    elif command == _READ_LINE:
        print(dev.readline())
    elif command == _DONE:
        _exit = True
    elif command in gcode_shortcuts:
        await handle_gcode_shortcut(dev, command)
    else:
        try:
            dev.write(f"{command}\n".encode())
            print(await message_return(dev))
        except TypeError:
            print("Invalid input.")
    return _exit


async def _main() -> None:
    tc_name = (
        subprocess.check_output(["find", "/dev/", "-name", "*thermocycler*"])
        .decode()
        .strip()
    )
    if not tc_name:
        print("Thermocycler not found. Exiting.")
        return
    dev = Serial(f"{tc_name}", 9600, timeout=2)
    _exit = False
    while not _exit:
        _exit = await comms_loop(dev)


if __name__ == "__main__":
    asyncio.run(_main())



"""Load lid stack command request, result, and implementation models."""

from __future__ import annotations
from pydantic import BaseModel, Field
from typing import TYPE_CHECKING, Optional, Type, List, Any
from pydantic.json_schema import SkipJsonSchema
from typing_extensions import Literal

from opentrons_shared_data.labware.labware_definition import LabwareDefinition

from ..errors import LabwareIsNotAllowedInLocationError, ProtocolEngineError
from ..resources import fixture_validation, labware_validation
from ..types import (
    LoadableLabwareLocation,
    SYSTEM_LOCATION,
    OnLabwareLocation,
    DeckSlotLocation,
    AddressableAreaLocation,
    LabwareLocationSequence,
    OnLabwareLocationSequenceComponent,
)

from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ..errors.error_occurrence import ErrorOccurrence
from ..state.update_types import StateUpdate

if TYPE_CHECKING:
    from ..state.state import StateView
    from ..execution import LoadedLabwareData, EquipmentHandler


LoadLidStackCommandType = Literal["loadLidStack"]


def _remove_default(s: dict[str, Any]) -> None:
    s.pop("default", None)


_LID_STACK_PE_LABWARE = "protocol_engine_lid_stack_object"
_LID_STACK_PE_NAMESPACE = "opentrons"
_LID_STACK_PE_VERSION = 1


class LoadLidStackParams(BaseModel):
    """Payload required to load a lid stack onto a location."""

    location: LoadableLabwareLocation = Field(
        ...,
        description="Location the lid stack should be loaded into.",
    )
    loadName: str = Field(
        ...,
        description="Name used to reference a lid labware definition.",
    )
    namespace: str = Field(
        ...,
        description="The namespace the lid labware definition belongs to.",
    )
    version: int = Field(
        ...,
        description="The lid labware definition version.",
    )
    stackLabwareId: str | SkipJsonSchema[None] = Field(
        None,
        description="An optional ID to assign to the lid stack labware object created."
        "If None, an ID will be generated.",
        json_schema_extra=_remove_default,
    )
    labwareIds: List[str] | SkipJsonSchema[None] = Field(
        None,
        description="An optional list of IDs to assign to the lids in the stack."
        "If None, an ID will be generated.",
        json_schema_extra=_remove_default,
    )
    quantity: int = Field(
        ...,
        description="The quantity of lids to load.",
    )


class LoadLidStackResult(BaseModel):
    """Result data from the execution of a LoadLidStack command."""

    stackLabwareId: str = Field(
        ...,
        description="An ID to reference the lid stack labware object created.",
    )
    labwareIds: List[str] = Field(
        ...,
        description="A list of lid labware IDs to reference the lids in this stack by. The first ID is the bottom of the stack.",
    )
    definition: LabwareDefinition | None = Field(
        ...,
        description="The full definition data for this lid labware.",
    )
    location: LoadableLabwareLocation = Field(
        ..., description="The Location that the stack of lid labware has been loaded."
    )
    lidStackDefinition: LabwareDefinition | None = Field(
        None,
        description="The definition of the lid stack object. Optional for backwards-compatibility.",
    )
    stackLocationSequence: LabwareLocationSequence | None = Field(
        None,
        description="The location sequence for the lid stack labware object created.",
    )
    locationSequences: List[LabwareLocationSequence] | None = Field(
        None,
        description="The location sequences for the lids just loaded into the stack. These are in the same order as labwareIds.",
    )


class LoadLidStackImplementation(
    AbstractCommandImpl[LoadLidStackParams, SuccessData[LoadLidStackResult]]
):
    """Load lid stack command implementation."""

    def __init__(
        self, equipment: EquipmentHandler, state_view: StateView, **kwargs: object
    ) -> None:
        self._equipment = equipment
        self._state_view = state_view

    def _validate_location(self, params: LoadLidStackParams) -> LoadableLabwareLocation:
        if isinstance(params.location, AddressableAreaLocation):
            area_name = params.location.addressableAreaName
            if not (
                fixture_validation.is_deck_slot(params.location.addressableAreaName)
                or fixture_validation.is_abs_reader(params.location.addressableAreaName)
            ):
                raise LabwareIsNotAllowedInLocationError(
                    f"Cannot load {params.loadName} onto addressable area {area_name}"
                )
            self._state_view.addressable_areas.raise_if_area_not_in_deck_configuration(
                area_name
            )
        elif isinstance(params.location, DeckSlotLocation):
            self._state_view.addressable_areas.raise_if_area_not_in_deck_configuration(
                params.location.slotName.id
            )
        if params.quantity <= 0 and params.location != SYSTEM_LOCATION:
            raise ProtocolEngineError(
                message="Lid Stack Labware Object with quantity 0 must be loaded onto System Location."
            )
        return self._state_view.geometry.ensure_location_not_occupied(params.location)

    def _format_results(
        self,
        verified_location: LoadableLabwareLocation,
        lid_stack_object: LoadedLabwareData,
        loaded_lid_labwares: list[LoadedLabwareData],
        lid_labware_definition: LabwareDefinition | None,
    ) -> SuccessData[LoadLidStackResult]:
        stack_location_sequence = (
            self._state_view.geometry.get_predicted_location_sequence(verified_location)
        )
        loaded_lid_locations_by_id: dict[str, OnLabwareLocation] = {}
        loaded_lid_ids_ordered: list[str] = []
        loaded_lid_location_sequences_ordered: list[LabwareLocationSequence] = []
        lid_location_sequence_accumulated: LabwareLocationSequence = [
            OnLabwareLocationSequenceComponent(
                labwareId=lid_stack_object.labware_id, lidId=None
            )
        ] + stack_location_sequence
        load_location = OnLabwareLocation(labwareId=lid_stack_object.labware_id)
        last_lid_id: str | None = None
        for loaded_lid in loaded_lid_labwares:
            loaded_lid_locations_by_id[loaded_lid.labware_id] = load_location
            loaded_lid_ids_ordered.append(loaded_lid.labware_id)
            if last_lid_id is None:
                last_lid_id = loaded_lid.labware_id
            else:
                lid_location_sequence_accumulated = [
                    OnLabwareLocationSequenceComponent(
                        labwareId=last_lid_id, lidId=None
                    )
                ] + lid_location_sequence_accumulated
                last_lid_id = loaded_lid.labware_id
            loaded_lid_location_sequences_ordered.append(
                [loc for loc in lid_location_sequence_accumulated]
            )
            load_location = OnLabwareLocation(labwareId=loaded_lid.labware_id)

        state_update = StateUpdate()
        state_update.set_loaded_lid_stack(
            stack_id=lid_stack_object.labware_id,
            stack_object_definition=lid_stack_object.definition,
            stack_location=verified_location,
            locations=loaded_lid_locations_by_id,
            labware_definition=lid_labware_definition,
        )

        return SuccessData(
            public=LoadLidStackResult(
                stackLabwareId=lid_stack_object.labware_id,
                lidStackDefinition=lid_stack_object.definition,
                labwareIds=loaded_lid_ids_ordered,
                definition=lid_labware_definition,
                location=verified_location,
                stackLocationSequence=stack_location_sequence,
                locationSequences=loaded_lid_location_sequences_ordered,
            ),
            state_update=state_update,
        )

    async def execute(
        self, params: LoadLidStackParams
    ) -> SuccessData[LoadLidStackResult]:
        """Load definition and calibration data necessary for a lid stack."""
        verified_location = self._validate_location(params)

        lid_stack_object = await self._equipment.load_labware(
            load_name=_LID_STACK_PE_LABWARE,
            namespace=_LID_STACK_PE_NAMESPACE,
            version=_LID_STACK_PE_VERSION,
            location=verified_location,
            labware_id=params.stackLabwareId,
        )

        if not labware_validation.validate_definition_is_system(
            lid_stack_object.definition
        ):
            raise ProtocolEngineError(
                message="Lid Stack Labware Object Labware Definition does not contain required allowed role 'system'."
            )

        loaded_lid_labwares: list[LoadedLabwareData] = []
        lid_labware_definition: LabwareDefinition | None = None

        if params.quantity > 0:
            loaded_lid_labwares = await self._equipment.load_lids(
                load_name=params.loadName,
                namespace=params.namespace,
                version=params.version,
                location=OnLabwareLocation(labwareId=lid_stack_object.labware_id),
                quantity=params.quantity,
                labware_ids=params.labwareIds,
            )
            lid_labware_definition = loaded_lid_labwares[-1].definition
            if isinstance(verified_location, OnLabwareLocation):
                self._state_view.labware.raise_if_labware_cannot_be_stacked(
                    top_labware_definition=lid_labware_definition,
                    bottom_labware_id=verified_location.labwareId,
                )

        return self._format_results(
            verified_location,
            lid_stack_object,
            loaded_lid_labwares,
            lid_labware_definition,
        )


class LoadLidStack(
    BaseCommand[LoadLidStackParams, LoadLidStackResult, ErrorOccurrence]
):
    """Load lid stack command resource model."""

    commandType: LoadLidStackCommandType = "loadLidStack"
    params: LoadLidStackParams
    result: Optional[LoadLidStackResult] = None

    _ImplementationCls: Type[LoadLidStackImplementation] = LoadLidStackImplementation


class LoadLidStackCreate(BaseCommandCreate[LoadLidStackParams]):
    """Load lid stack command creation request."""

    commandType: LoadLidStackCommandType = "loadLidStack"
    params: LoadLidStackParams

    _CommandCls: Type[LoadLidStack] = LoadLidStack



from datetime import datetime

from dataclasses import dataclass
from typing import List, Optional
from opentrons.calibration_storage.ot3.models.v1 import CalibrationStatus
from opentrons.calibration_storage.ot3.module_offset import (
    get_module_offset,
    load_all_module_offsets,
    save_module_calibration,
)
from opentrons.calibration_storage.types import SourceType
from opentrons.hardware_control.modules.types import ModuleType
from opentrons.hardware_control.types import OT3Mount


from opentrons.types import Point


@dataclass
class ModuleCalibrationOffset:
    """Class to store module offset calibration data."""

    offset: Point
    module_id: str
    module: ModuleType
    source: SourceType
    status: CalibrationStatus
    slot: str
    mount: Optional[OT3Mount] = None
    instrument_id: Optional[str] = None
    last_modified: Optional[datetime] = None


def load_module_calibration_offset(
    module_type: ModuleType,
    module_id: str,
) -> Optional[ModuleCalibrationOffset]:
    """Loads the calibration offset for a module."""
    module_offset_data = get_module_offset(module_type, module_id)
    if not module_offset_data:
        return None
    return ModuleCalibrationOffset(
        module=module_type,
        module_id=module_id,
        slot=module_offset_data.slot,
        mount=module_offset_data.mount,
        offset=module_offset_data.offset,
        last_modified=module_offset_data.lastModified,
        instrument_id=module_offset_data.instrument_id,
        source=module_offset_data.source,
        status=CalibrationStatus(
            markedAt=module_offset_data.status.markedAt,
            markedBad=module_offset_data.status.markedBad,
            source=module_offset_data.status.source,
        ),
    )


def save_module_calibration_offset(
    offset: Point,
    mount: OT3Mount,
    slot: str,
    module: ModuleType,
    module_id: str,
    instrument_id: Optional[str] = None,
) -> None:
    """Save the calibration offset for a given module."""
    save_module_calibration(offset, mount, slot, module, module_id, instrument_id)


def load_all_module_calibrations() -> List[ModuleCalibrationOffset]:
    """Loads all the module calibration stored on the robot."""
    module_calibrations: List[ModuleCalibrationOffset] = []
    module_offset_data = load_all_module_offsets()
    for data in module_offset_data:
        module_calibrations.append(
            ModuleCalibrationOffset(
                slot=data.slot,
                module=data.module,
                module_id=data.module_id,
                mount=data.mount,
                offset=data.offset,
                last_modified=data.lastModified,
                instrument_id=data.instrument_id,
                source=data.source,
                status=CalibrationStatus(
                    markedAt=data.status.markedAt,
                    markedBad=data.status.markedBad,
                    source=data.status.source,
                ),
            )
        )
    return module_calibrations



"""Drop tip command request, result, and implementation models."""

from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type, Any

from pydantic import Field
from pydantic.json_schema import SkipJsonSchema

from typing_extensions import Literal

from opentrons.protocol_engine.errors.exceptions import TipAttachedError
from opentrons.protocol_engine.resources.model_utils import ModelUtils

from ..state.update_types import StateUpdate
from ..types import DropTipWellLocation
from .pipetting_common import (
    PipetteIdMixin,
    TipPhysicallyAttachedError,
)
from .movement_common import (
    DestinationPositionResult,
    move_to_well,
    StallOrCollisionError,
)
from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    DefinedErrorData,
    SuccessData,
)
from ..errors.error_occurrence import ErrorOccurrence

if TYPE_CHECKING:
    from ..state.state import StateView
    from ..execution import MovementHandler, TipHandler

from opentrons.hardware_control.types import TipScrapeType

DropTipCommandType = Literal["dropTip"]


def _remove_default(s: dict[str, Any]) -> None:
    s.pop("default", None)


class DropTipParams(PipetteIdMixin):
    """Payload required to drop a tip in a specific well."""

    labwareId: str = Field(..., description="Identifier of labware to use.")
    wellName: str = Field(..., description="Name of well to use in labware.")
    wellLocation: DropTipWellLocation = Field(
        default_factory=DropTipWellLocation,
        description="Relative well location at which to drop the tip.",
    )
    homeAfter: bool | SkipJsonSchema[None] = Field(
        None,
        description=(
            "Whether to home this pipette's plunger after dropping the tip."
            " You should normally leave this unspecified to let the robot choose"
            " a safe default depending on its hardware."
        ),
        json_schema_extra=_remove_default,
    )
    alternateDropLocation: bool | SkipJsonSchema[None] = Field(
        False,
        description=(
            "Whether to alternate location where tip is dropped within the labware."
            " If True, this command will ignore the wellLocation provided and alternate"
            " between dropping tips at two predetermined locations inside the specified"
            " labware well."
            " If False, the tip will be dropped at the top center of the well."
        ),
        json_schema_extra=_remove_default,
    )
    scrape_tips: bool | SkipJsonSchema[None] = Field(
        False,
        description=(
            "Whether or not to scrape off the tips with the ejector all the way down."
            " If True, and the target location is a tip rack well, it will move the pipette."
            " Towards the center of the tip rack with the ejector in the 'drop_tip' position."
            " If False, no horizontal movement will occur."
        ),
        json_schema_extra=_remove_default,
    )


class DropTipResult(DestinationPositionResult):
    """Result data from the execution of a DropTip command."""

    pass


_ExecuteReturn = (
    SuccessData[DropTipResult]
    | DefinedErrorData[TipPhysicallyAttachedError]
    | DefinedErrorData[StallOrCollisionError]
)


class DropTipImplementation(AbstractCommandImpl[DropTipParams, _ExecuteReturn]):
    """Drop tip command implementation."""

    def __init__(
        self,
        state_view: StateView,
        tip_handler: TipHandler,
        movement: MovementHandler,
        model_utils: ModelUtils,
        **kwargs: object,
    ) -> None:
        self._state_view = state_view
        self._tip_handler = tip_handler
        self._movement_handler = movement
        self._model_utils = model_utils

    async def execute(self, params: DropTipParams) -> _ExecuteReturn:
        """Move to and drop a tip using the requested pipette."""
        pipette_id = params.pipetteId
        labware_id = params.labwareId
        well_name = params.wellName
        home_after = params.homeAfter

        if params.alternateDropLocation:
            well_location = self._state_view.geometry.get_next_tip_drop_location(
                labware_id=labware_id,
                well_name=well_name,
                pipette_id=pipette_id,
            )
        else:
            well_location = params.wellLocation

        is_partially_configured = self._state_view.pipettes.get_is_partially_configured(
            pipette_id=pipette_id
        )
        tip_drop_location = self._state_view.geometry.get_checked_tip_drop_location(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_location=well_location,
            partially_configured=is_partially_configured,
        )

        move_result = await move_to_well(
            movement=self._movement_handler,
            model_utils=self._model_utils,
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=tip_drop_location,
        )
        if isinstance(move_result, DefinedErrorData):
            return move_result

        scrape_type = TipScrapeType.NONE
        if (
            params.scrape_tips
            and self._state_view.geometry._labware.get_definition(
                labware_id
            ).parameters.isTiprack
        ):
            if int("".join(filter(str.isdigit, well_name))) <= 6:
                scrape_type = TipScrapeType.RIGHT_ONE_COL
            else:
                scrape_type = TipScrapeType.LEFT_ONE_COL
        try:
            await self._tip_handler.drop_tip(
                pipette_id=pipette_id, home_after=home_after, scrape_type=scrape_type
            )
        except TipAttachedError as exception:
            error = TipPhysicallyAttachedError(
                id=self._model_utils.generate_id(),
                createdAt=self._model_utils.get_timestamp(),
                wrappedErrors=[
                    ErrorOccurrence.from_failed(
                        id=self._model_utils.generate_id(),
                        createdAt=self._model_utils.get_timestamp(),
                        error=exception,
                    )
                ],
                errorInfo={
                    "retryLocation": (
                        move_result.public.position.x,
                        move_result.public.position.y,
                        move_result.public.position.z,
                    )
                },
            )
            return DefinedErrorData(
                public=error,
                state_update=StateUpdate.reduce(
                    StateUpdate(), move_result.state_update
                ).set_fluid_unknown(pipette_id=pipette_id),
                state_update_if_false_positive=move_result.state_update.update_pipette_tip_state(
                    pipette_id=params.pipetteId, tip_geometry=None
                ),
            )
        else:
            return SuccessData(
                public=DropTipResult(position=move_result.public.position),
                state_update=move_result.state_update.set_fluid_unknown(
                    pipette_id=pipette_id
                ).update_pipette_tip_state(
                    pipette_id=params.pipetteId, tip_geometry=None
                ),
            )


class DropTip(
    BaseCommand[
        DropTipParams, DropTipResult, TipPhysicallyAttachedError | StallOrCollisionError
    ]
):
    """Drop tip command model."""

    commandType: DropTipCommandType = "dropTip"
    params: DropTipParams
    result: Optional[DropTipResult] = None

    _ImplementationCls: Type[DropTipImplementation] = DropTipImplementation


class DropTipCreate(BaseCommandCreate[DropTipParams]):
    """Drop tip command creation request model."""

    commandType: DropTipCommandType = "dropTip"
    params: DropTipParams

    _CommandCls: Type[DropTip] = DropTip



#!/usr/bin/env python3

"""Protocol Engine types that have to do with runtime parameters."""

from pathlib import Path
from typing import Optional, Mapping, Union, Literal, Dict, List
from pydantic import (
    BaseModel,
    Field,
    StrictBool,
    StrictFloat,
    StrictInt,
    StrictStr,
)


# TODO (spp, 2024-04-02): move all RTP types to runner
class RTPBase(BaseModel):
    """Parameters defined in a protocol."""

    displayName: StrictStr = Field(..., description="Display string for the parameter.")
    variableName: StrictStr = Field(
        ..., description="Python variable name of the parameter."
    )
    description: Optional[StrictStr] = Field(
        None, description="Detailed description of the parameter."
    )
    suffix: Optional[StrictStr] = Field(
        None,
        description="Units (like mL, mm/sec, etc) or a custom suffix for the parameter.",
    )


class NumberParameter(RTPBase):
    """An integer parameter defined in a protocol."""

    type: Literal["int", "float"] = Field(
        ..., description="String specifying whether the number is an int or float type."
    )
    min: Union[StrictInt, StrictFloat] = Field(
        ..., description="Minimum value that the number param is allowed to have."
    )
    max: Union[StrictInt, StrictFloat] = Field(
        ..., description="Maximum value that the number param is allowed to have."
    )
    value: Union[StrictInt, StrictFloat] = Field(
        ...,
        description="The value assigned to the parameter; if not supplied by the client, will be assigned the default value.",
    )
    default: Union[StrictInt, StrictFloat] = Field(
        ...,
        description="Default value of the parameter, to be used when there is no client-specified value.",
    )


class BooleanParameter(RTPBase):
    """A boolean parameter defined in a protocol."""

    type: Literal["bool"] = Field(
        default="bool", description="String specifying the type of this parameter"
    )
    value: StrictBool = Field(
        ...,
        description="The value assigned to the parameter; if not supplied by the client, will be assigned the default value.",
    )
    default: StrictBool = Field(
        ...,
        description="Default value of the parameter, to be used when there is no client-specified value.",
    )


class EnumChoice(BaseModel):
    """Components of choices used in RTP Enum Parameters."""

    displayName: StrictStr = Field(
        ..., description="Display string for the param's choice."
    )
    value: Union[StrictInt, StrictFloat, StrictStr] = Field(
        ..., description="Enum value of the param's choice."
    )


class EnumParameter(RTPBase):
    """A string enum defined in a protocol."""

    type: Literal["int", "float", "str"] = Field(
        ...,
        description="String specifying whether the parameter is an int or float or string type.",
    )
    choices: List[EnumChoice] = Field(
        ..., description="List of valid choices for this parameter."
    )
    value: Union[StrictInt, StrictFloat, StrictStr] = Field(
        ...,
        description="The value assigned to the parameter; if not supplied by the client, will be assigned the default value.",
    )
    default: Union[StrictInt, StrictFloat, StrictStr] = Field(
        ...,
        description="Default value of the parameter, to be used when there is no client-specified value.",
    )


class FileInfo(BaseModel):
    """A file UUID descriptor."""

    id: str = Field(
        ...,
        description="The UUID identifier of the file stored on the robot.",
    )
    name: str = Field(..., description="Name of the file, including the extension.")


class CSVParameter(RTPBase):
    """A CSV file parameter defined in a protocol."""

    type: Literal["csv_file"] = Field(
        default="csv_file", description="String specifying the type of this parameter"
    )
    file: Optional[FileInfo] = Field(
        default=None,
        description="ID of the CSV file stored on the robot; to be used for fetching the CSV file."
        " For local analysis this will most likely be empty.",
    )


RunTimeParameter = Union[NumberParameter, EnumParameter, BooleanParameter, CSVParameter]

PrimitiveRunTimeParamValuesType = Mapping[
    StrictStr, Union[StrictInt, StrictFloat, StrictBool, StrictStr]
]  # update value types as more RTP types are added

CSVRunTimeParamFilesType = Mapping[StrictStr, StrictStr]
CSVRuntimeParamPaths = Dict[str, Path]



"""Helper functions for liquid-level related calculations inside a given frustum."""
from typing import List, Tuple
from numpy import pi, iscomplex, roots, real
from math import isclose

from ..errors.exceptions import InvalidLiquidHeightFound

from opentrons.protocol_engine.types.liquid_level_detection import (
    LiquidTrackingType,
    SimulatedProbeResult,
)
from opentrons_shared_data.labware.labware_definition import (
    InnerWellGeometry,
    WellSegment,
    SphericalSegment,
    ConicalFrustum,
    CuboidalFrustum,
    SquaredConeSegment,
)


def _reject_unacceptable_heights(
    potential_heights: List[float], max_height: float
) -> float:
    """Reject any solutions to a polynomial equation that cannot be the height of a frustum."""
    valid_heights: List[float] = []
    for root in potential_heights:
        # reject any heights that are negative or greater than the max height
        if not iscomplex(root):
            # take only the real component of the root and round to 4 decimal places
            rounded_root = round(real(root), 4)
            if (rounded_root <= max_height) and (rounded_root >= 0):
                if not any([isclose(rounded_root, height) for height in valid_heights]):
                    valid_heights.append(rounded_root)
    if len(valid_heights) != 1:
        raise InvalidLiquidHeightFound(
            message="Unable to estimate valid liquid height from volume."
        )
    return valid_heights[0]


def _cross_section_area_circular(diameter: float) -> float:
    """Get the area of a circular cross-section."""
    radius = diameter / 2
    return pi * (radius**2)


def _cross_section_area_rectangular(x_dimension: float, y_dimension: float) -> float:
    """Get the area of a rectangular cross-section."""
    return x_dimension * y_dimension


def _rectangular_frustum_polynomial_roots(
    bottom_length: float,
    bottom_width: float,
    top_length: float,
    top_width: float,
    total_frustum_height: float,
) -> Tuple[float, float, float]:
    """Polynomial representation of the volume of a rectangular frustum."""
    # roots of the polynomial with shape ax^3 + bx^2 + cx
    a = (
        (top_length - bottom_length)
        * (top_width - bottom_width)
        / (3 * total_frustum_height**2)
    )
    b = (
        (bottom_length * (top_width - bottom_width))
        + (bottom_width * (top_length - bottom_length))
    ) / (2 * total_frustum_height)
    c = bottom_length * bottom_width
    return a, b, c


def _circular_frustum_polynomial_roots(
    bottom_radius: float,
    top_radius: float,
    total_frustum_height: float,
) -> Tuple[float, float, float]:
    """Polynomial representation of the volume of a circular frustum."""
    # roots of the polynomial with shape ax^3 + bx^2 + cx
    a = pi * ((top_radius - bottom_radius) ** 2) / (3 * total_frustum_height**2)
    b = pi * bottom_radius * (top_radius - bottom_radius) / total_frustum_height
    c = pi * bottom_radius**2
    return a, b, c


def _volume_from_height_circular(
    target_height: float, segment: ConicalFrustum
) -> float:
    return segment.volume_from_height_circular(
        top_radius=segment.topDiameter / 2,
        bottom_radius=segment.bottomDiameter / 2,
        target_height=target_height,
        total_height=segment.topHeight - segment.bottomHeight,
    )


def _volume_from_height_rectangular(
    target_height: float,
    total_frustum_height: float,
    bottom_length: float,
    bottom_width: float,
    top_length: float,
    top_width: float,
) -> float:
    """Find the volume given a height within a rectangular frustum."""
    a, b, c = _rectangular_frustum_polynomial_roots(
        bottom_length=bottom_length,
        bottom_width=bottom_width,
        top_length=top_length,
        top_width=top_width,
        total_frustum_height=total_frustum_height,
    )
    volume = a * (target_height**3) + b * (target_height**2) + c * target_height
    return volume


def _volume_from_height_spherical(
    target_height: float,
    radius_of_curvature: float,
) -> float:
    """Find the volume given a height within a spherical frustum."""
    volume = (
        (1 / 3) * pi * (target_height**2) * (3 * radius_of_curvature - target_height)
    )
    return volume


def _volume_from_height_squared_cone(
    target_height: float, segment: SquaredConeSegment
) -> float:
    """Find the volume given a height within a squared cone segment."""
    heights = segment.height_to_volume_table.keys()
    best_fit_height = min(heights, key=lambda x: abs(x - target_height))
    return segment.height_to_volume_table[best_fit_height]


def _height_from_volume_circular(
    target_volume: float, segment: ConicalFrustum
) -> float:
    """Find the height given a volume within a squared cone segment."""
    return segment.height_from_volume_search(target_volume)


def _height_from_volume_rectangular(
    volume: float,
    total_frustum_height: float,
    bottom_length: float,
    bottom_width: float,
    top_length: float,
    top_width: float,
) -> float:
    """Find the height given a volume within a rectangular frustum."""
    a, b, c = _rectangular_frustum_polynomial_roots(
        bottom_length=bottom_length,
        bottom_width=bottom_width,
        top_length=top_length,
        top_width=top_width,
        total_frustum_height=total_frustum_height,
    )
    d = volume * -1
    x_intercept_roots = (a, b, c, d)

    height_from_volume_roots = roots(x_intercept_roots)
    height = _reject_unacceptable_heights(
        potential_heights=list(height_from_volume_roots),
        max_height=total_frustum_height,
    )
    return height


def _height_from_volume_spherical(
    volume: float,
    radius_of_curvature: float,
    total_frustum_height: float,
) -> float:
    """Find the height given a volume within a spherical frustum."""
    a = -1 * pi / 3
    b = pi * radius_of_curvature
    c = 0.0
    d = volume * -1
    x_intercept_roots = (a, b, c, d)

    height_from_volume_roots = roots(x_intercept_roots)
    height = _reject_unacceptable_heights(
        potential_heights=list(height_from_volume_roots),
        max_height=total_frustum_height,
    )
    return height


def _height_from_volume_squared_cone(
    target_volume: float, segment: SquaredConeSegment
) -> float:
    """Find the height given a volume within a squared cone segment."""
    volumes = segment.volume_to_height_table.keys()
    best_fit_volume = min(volumes, key=lambda x: abs(x - target_volume))
    return segment.volume_to_height_table[best_fit_volume]


def _get_segment_capacity(segment: WellSegment) -> float:
    section_height = segment.topHeight - segment.bottomHeight
    match segment:
        case SphericalSegment():
            return (
                _volume_from_height_spherical(
                    target_height=segment.topHeight,
                    radius_of_curvature=segment.radiusOfCurvature,
                )
                * segment.count
            )
        case CuboidalFrustum():
            return (
                _volume_from_height_rectangular(
                    target_height=section_height,
                    bottom_length=segment.bottomYDimension,
                    bottom_width=segment.bottomXDimension,
                    top_length=segment.topYDimension,
                    top_width=segment.topXDimension,
                    total_frustum_height=section_height,
                )
                * segment.count
            )
        case ConicalFrustum():
            return (
                _volume_from_height_circular(
                    target_height=section_height,
                    segment=segment,
                )
                * segment.count
            )
        case SquaredConeSegment():
            return (
                _volume_from_height_squared_cone(section_height, segment)
                * segment.count
            )
        case _:
            # TODO: implement volume calculations for truncated circular and rounded rectangular segments
            raise NotImplementedError(
                f"volume calculation for shape: {segment.shape} not yet implemented."
            )


def get_well_volumetric_capacity(
    well_geometry: InnerWellGeometry,
) -> List[Tuple[float, float]]:
    """Return the volumetric capacity of a well as a list of pairs relating segment heights to volumes."""
    #  [(top_height_0, section_0_volume), (top_height_1, section_1_volume), ...]
    well_volume = []

    # get the well segments sorted in ascending order
    sorted_well = sorted(well_geometry.sections, key=lambda section: section.topHeight)

    for segment in sorted_well:
        section_volume = _get_segment_capacity(segment)
        well_volume.append((segment.topHeight, section_volume))
    return well_volume


def height_at_volume_within_section(
    section: WellSegment,
    target_volume_relative: float,
    section_height: float,
) -> float:
    """Calculate a height within a bounded section according to geometry."""
    target_volume_relative = target_volume_relative / section.count
    match section:
        case SphericalSegment():
            return _height_from_volume_spherical(
                volume=target_volume_relative,
                total_frustum_height=section_height,
                radius_of_curvature=section.radiusOfCurvature,
            )
        case ConicalFrustum():
            return _height_from_volume_circular(target_volume_relative, section)
        case CuboidalFrustum():
            return _height_from_volume_rectangular(
                volume=target_volume_relative,
                total_frustum_height=section_height,
                bottom_width=section.bottomXDimension,
                bottom_length=section.bottomYDimension,
                top_width=section.topXDimension,
                top_length=section.topYDimension,
            )
        case SquaredConeSegment():
            return _height_from_volume_squared_cone(target_volume_relative, section)
        case _:
            raise NotImplementedError(
                "Height from volume calculation not yet implemented for this well shape."
            )


def volume_at_height_within_section(
    section: WellSegment,
    target_height_relative: float,
    section_height: float,
) -> float:
    """Calculate a volume within a bounded section according to geometry."""
    match section:
        case SphericalSegment():
            return (
                _volume_from_height_spherical(
                    target_height=target_height_relative,
                    radius_of_curvature=section.radiusOfCurvature,
                )
                * section.count
            )
        case ConicalFrustum():
            return (
                _volume_from_height_circular(
                    target_height=target_height_relative, segment=section
                )
                * section.count
            )
        case CuboidalFrustum():
            return (
                _volume_from_height_rectangular(
                    target_height=target_height_relative,
                    total_frustum_height=section_height,
                    bottom_width=section.bottomXDimension,
                    bottom_length=section.bottomYDimension,
                    top_width=section.topXDimension,
                    top_length=section.topYDimension,
                )
                * section.count
            )
        case SquaredConeSegment():
            return (
                _volume_from_height_squared_cone(target_height_relative, section)
                * section.count
            )
        case _:
            # TODO(cm): this would be the NEST-96 2uL wells referenced in EXEC-712
            # we need to input the math attached to that issue
            raise NotImplementedError(
                "Height from volume calculation not yet implemented for this well shape."
            )


def _find_volume_in_partial_frustum(
    sorted_well: List[WellSegment],
    target_height: float,
) -> float:
    """Look through a sorted list of frusta for a target height, and find the volume at that height."""
    for segment in sorted_well:
        if segment.bottomHeight <= target_height <= segment.topHeight:
            relative_target_height = target_height - segment.bottomHeight
            section_height = segment.topHeight - segment.bottomHeight
            return volume_at_height_within_section(
                section=segment,
                target_height_relative=relative_target_height,
                section_height=section_height,
            )
    # if we've looked through all sections and can't find the target volume, raise an error
    # this code should never be reached- an error should be raised by find_volume_at_well_height
    raise InvalidLiquidHeightFound(
        f"Target height {target_height} mm exceeds the well height."
    )


def find_volume_at_well_height(
    target_height: LiquidTrackingType,
    well_geometry: InnerWellGeometry,
) -> LiquidTrackingType:
    """Find the volume within a well, at a known height."""
    # comparisons with SimulatedProbeResult objects aren't meaningful, just return
    if isinstance(target_height, SimulatedProbeResult):
        return target_height
    volumetric_capacity = get_well_volumetric_capacity(well_geometry)
    max_height = volumetric_capacity[-1][0]
    if target_height < 0 or target_height > max_height:
        raise InvalidLiquidHeightFound(
            f"Invalid target height {target_height} mm; max well height is {max_height} mm."
        )
    # volumes in volumetric_capacity are relative to each frustum,
    # so we have to find the volume of all the full sections enclosed
    # beneath the target height
    closed_section_volume = 0.0
    for boundary_height, section_volume in volumetric_capacity:
        if boundary_height > target_height:
            break
        closed_section_volume += section_volume
        # if target height is a boundary cross-section, we already know the volume
        if target_height == boundary_height:
            return closed_section_volume
    # find the section the target height is in and compute the volume

    sorted_well = sorted(well_geometry.sections, key=lambda section: section.topHeight)
    partial_volume = _find_volume_in_partial_frustum(
        sorted_well=sorted_well,
        target_height=target_height,
    )
    return partial_volume + closed_section_volume


def _find_height_in_partial_frustum(
    sorted_well: List[WellSegment],
    volumetric_capacity: List[Tuple[float, float]],
    target_volume: float,
) -> float:
    """Look through a sorted list of frusta for a target volume, and find the height at that volume."""
    bottom_section_volume = 0.0
    if target_volume == 0.0:
        return 0.0
    for section, capacity in zip(sorted_well, volumetric_capacity):
        section_top_height, section_volume = capacity
        if target_volume == section_volume + bottom_section_volume:
            return section_top_height
        if (
            bottom_section_volume
            <= target_volume
            <= (bottom_section_volume + section_volume)
        ):
            relative_target_volume = target_volume - bottom_section_volume
            section_height = section.topHeight - section.bottomHeight
            partial_height = height_at_volume_within_section(
                section=section,
                target_volume_relative=relative_target_volume,
                section_height=section_height,
            )
            return partial_height + section.bottomHeight
        # bottom section volume should always be the volume enclosed in the previously
        # viewed section
        bottom_section_volume += section_volume

    # if we finish looping through the whole well, bottom_section will be the well's volume
    total_well_volume = bottom_section_volume
    # if we've looked through all sections and can't find the target volume, raise an error
    # also this code should never be reached bc an invalid target volume should be changed
    # by find_height_at_well_volume
    raise InvalidLiquidHeightFound(
        f"Target volume {target_volume} uL exceeds the well volume {total_well_volume} uL."
    )


def find_height_at_well_volume(
    target_volume: LiquidTrackingType,
    well_geometry: InnerWellGeometry,
) -> LiquidTrackingType:
    """Find the height within a well, at a known volume."""
    # comparisons with SimulatedProbeResult objects aren't meaningful, just
    # return if we have one of those
    if isinstance(target_volume, SimulatedProbeResult):
        return target_volume

    volumetric_capacity = get_well_volumetric_capacity(well_geometry)
    max_volume = sum(row[1] for row in volumetric_capacity)

    if target_volume < 0:
        target_volume = 0
    elif target_volume > max_volume:
        target_volume = max_volume
    sorted_well = sorted(well_geometry.sections, key=lambda section: section.topHeight)
    # find the section the target volume is in and compute the height
    return _find_height_in_partial_frustum(
        sorted_well=sorted_well,
        volumetric_capacity=volumetric_capacity,
        target_volume=target_volume,
    )



"""Wrappers for Protocol API v2 execution pipeline."""
import asyncio
from typing import Dict, Iterable, Optional, cast

from anyio import to_thread

from opentrons_shared_data.labware.types import (
    LabwareDefinition as LabwareDefinitionTypedDict,
)
from opentrons_shared_data.labware.labware_definition import LabwareDefinition

from opentrons.calibration_storage.helpers import uri_from_details
from opentrons.hardware_control import HardwareControlAPI
from opentrons.legacy_broker import LegacyBroker
from opentrons.protocol_engine import ProtocolEngine
from opentrons.protocol_engine.types import (
    PrimitiveRunTimeParamValuesType,
    CSVRuntimeParamPaths,
)
from opentrons.protocol_reader import ProtocolSource, ProtocolFileRole
from opentrons.util.broker import Broker

from opentrons.protocol_api import (
    ProtocolContext,
    ParameterContext,
    create_protocol_context,
    Parameters,
)
from opentrons.protocol_api.core.engine import ENGINE_CORE_API_VERSION
from opentrons.protocol_api.core.legacy.load_info import LoadInfo

from opentrons.protocols.parse import PythonParseMode, parse
from opentrons.protocols.execution.execute import run_protocol
from opentrons.protocols.execution.execute_python import exec_add_parameters
from opentrons.protocols.types import Protocol, PythonProtocol


# The earliest Python Protocol API version ("apiLevel") where the protocol's simulation
# and execution will be handled by Protocol Engine, rather than the previous direct hardware calls from protocol api.
#
# Note that even when simulation and execution are handled by the legacy machinery,
# Protocol Engine still has some involvement for analyzing the simulation and
# monitoring the execution.
LEGACY_PYTHON_API_VERSION_CUTOFF = ENGINE_CORE_API_VERSION


# The earliest JSON protocol schema version where the protocol is executed directly by
# Protocol Engine, rather than going through Python Protocol API v2.
LEGACY_JSON_SCHEMA_VERSION_CUTOFF = 6


class PythonAndLegacyFileReader:
    """Interface to read Protocol API v2 protocols prior to execution."""

    @staticmethod
    def read(
        protocol_source: ProtocolSource,
        labware_definitions: Iterable[LabwareDefinition],
        python_parse_mode: PythonParseMode,
    ) -> Protocol:
        """Read a PAPIv2 protocol into a data structure."""
        protocol_file_path = protocol_source.main_file
        protocol_contents = protocol_file_path.read_text(encoding="utf-8")
        extra_labware: Dict[str, LabwareDefinitionTypedDict] = {
            uri_from_details(
                namespace=lw.namespace,
                load_name=lw.parameters.loadName,
                version=lw.version,
            ): cast(LabwareDefinitionTypedDict, lw.model_dump(exclude_none=True))
            for lw in labware_definitions
        }
        data_file_paths = [
            data_file.path
            for data_file in protocol_source.files
            if data_file.role == ProtocolFileRole.DATA
        ]

        return parse(
            protocol_file=protocol_contents,
            filename=protocol_file_path.name,
            extra_labware=extra_labware,
            extra_data={
                data_path.name: data_path.read_bytes() for data_path in data_file_paths
            },
            python_parse_mode=python_parse_mode,
        )


class ProtocolContextCreator:
    """Interface to construct Protocol API v2 contexts."""

    _USE_SIMULATING_CORE = False

    def __init__(
        self,
        hardware_api: HardwareControlAPI,
        protocol_engine: ProtocolEngine,
    ) -> None:
        """Prepare the LegacyContextCreator.

        Args:
            hardware_api: The hardware control interface.
                Will be wrapped in a `SynchronousAdapter`.
                May be real hardware or a simulator.
            protocol_engine: Interface for the context to load labware offsets.
        """
        self._hardware_api = hardware_api
        self._protocol_engine = protocol_engine

    def create(
        self,
        protocol: Protocol,
        broker: Optional[LegacyBroker],
        equipment_broker: Optional[Broker[LoadInfo]],
    ) -> ProtocolContext:
        """Create a Protocol API v2 context."""
        extra_labware = (
            protocol.extra_labware if isinstance(protocol, PythonProtocol) else None
        )

        bundled_data = (
            protocol.bundled_data if isinstance(protocol, PythonProtocol) else None
        )

        return create_protocol_context(
            api_version=protocol.api_level,
            hardware_api=self._hardware_api,
            deck_type=self._protocol_engine.state_view.config.deck_type.value,
            protocol_engine=self._protocol_engine,
            protocol_engine_loop=asyncio.get_running_loop(),
            broker=broker,
            equipment_broker=equipment_broker,
            extra_labware=extra_labware,
            use_simulating_core=self._USE_SIMULATING_CORE,
            bundled_data=bundled_data,
        )


class SimulatingContextCreator(ProtocolContextCreator):
    """Interface to construct PAPIv2 contexts using simulating implementations.

    Avoids some calls to the hardware API for performance.
    See `opentrons.protocols.context.simulator`.
    """

    _USE_SIMULATING_CORE = True


class PythonProtocolExecutor:
    """Interface to execute Protocol API v2 protocols in a child thread."""

    @staticmethod
    async def execute(
        protocol: Protocol,
        context: ProtocolContext,
        run_time_parameters_with_overrides: Optional[Parameters],
    ) -> None:
        """Execute a PAPIv2 protocol with a given ProtocolContext in a child thread."""
        await to_thread.run_sync(
            run_protocol,
            protocol,
            context,
            run_time_parameters_with_overrides,
        )

    @staticmethod
    def extract_run_parameters(
        protocol: PythonProtocol,
        parameter_context: ParameterContext,
        run_time_param_overrides: Optional[PrimitiveRunTimeParamValuesType],
        run_time_param_file_overrides: Optional[CSVRuntimeParamPaths],
    ) -> Optional[Parameters]:
        """Extract the parameters defined in the protocol, overridden with values for the run."""
        return exec_add_parameters(
            protocol=protocol,
            parameter_context=parameter_context,
            run_time_param_overrides=run_time_param_overrides,
            run_time_param_file_overrides=run_time_param_file_overrides,
        )



"""Configure for volume command request, result, and implementation models."""
from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type, Any

from pydantic import BaseModel, Field
from pydantic.json_schema import SkipJsonSchema
from typing_extensions import Literal

from .pipetting_common import PipetteIdMixin
from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ..errors.error_occurrence import ErrorOccurrence
from ..state.update_types import StateUpdate

if TYPE_CHECKING:
    from ..execution import EquipmentHandler


ConfigureForVolumeCommandType = Literal["configureForVolume"]


def _remove_default(s: dict[str, Any]) -> None:
    s.pop("default", None)


class ConfigureForVolumeParams(PipetteIdMixin):
    """Parameters required to configure volume for a specific pipette."""

    volume: float = Field(
        ...,
        description="Amount of liquid in uL. Must be at least 0 and no greater "
        "than a pipette-specific maximum volume.",
        ge=0,
    )
    tipOverlapNotAfterVersion: str | SkipJsonSchema[None] = Field(
        None,
        description="A version of tip overlap data to not exceed. The highest-versioned "
        "tip overlap data that does not exceed this version will be used. Versions are "
        "expressed as vN where N is an integer, counting up from v0. If None, the current "
        "highest version will be used.",
        json_schema_extra=_remove_default,
    )


class ConfigureForVolumeResult(BaseModel):
    """Result data from execution of an ConfigureForVolume command."""

    pass


class ConfigureForVolumeImplementation(
    AbstractCommandImpl[
        ConfigureForVolumeParams,
        SuccessData[ConfigureForVolumeResult],
    ]
):
    """Configure for volume command implementation."""

    def __init__(self, equipment: EquipmentHandler, **kwargs: object) -> None:
        self._equipment = equipment

    async def execute(
        self, params: ConfigureForVolumeParams
    ) -> SuccessData[ConfigureForVolumeResult]:
        """Check that requested pipette can be configured for the given volume."""
        pipette_result = await self._equipment.configure_for_volume(
            pipette_id=params.pipetteId,
            volume=params.volume,
            tip_overlap_version=params.tipOverlapNotAfterVersion,
        )

        state_update = StateUpdate()
        state_update.update_pipette_config(
            pipette_id=pipette_result.pipette_id,
            config=pipette_result.static_config,
            serial_number=pipette_result.serial_number,
        )
        state_update.set_pipette_ready_to_aspirate(
            pipette_result.pipette_id, ready_to_aspirate=False
        )

        return SuccessData(
            public=ConfigureForVolumeResult(),
            state_update=state_update,
        )


class ConfigureForVolume(
    BaseCommand[ConfigureForVolumeParams, ConfigureForVolumeResult, ErrorOccurrence]
):
    """Configure for volume command model."""

    commandType: ConfigureForVolumeCommandType = "configureForVolume"
    params: ConfigureForVolumeParams
    result: Optional[ConfigureForVolumeResult] = None

    _ImplementationCls: Type[
        ConfigureForVolumeImplementation
    ] = ConfigureForVolumeImplementation


class ConfigureForVolumeCreate(BaseCommandCreate[ConfigureForVolumeParams]):
    """Configure for volume command creation request model."""

    commandType: ConfigureForVolumeCommandType = "configureForVolume"
    params: ConfigureForVolumeParams

    _CommandCls: Type[ConfigureForVolume] = ConfigureForVolume



import logging
from typing import Dict, List, TYPE_CHECKING, Union
from opentrons.hardware_control.modules.types import ThermocyclerStep
from opentrons.protocol_api import (
    ProtocolContext,
    MagneticModuleContext,
    TemperatureModuleContext,
    ModuleContext,
    ThermocyclerContext,
)
from opentrons_shared_data.protocol.types import (
    MagneticModuleCommand,
    TemperatureModuleCommand,
    ThermocyclerCommand,
)
from .execute_json_v3 import _delay, _move_to_slot
from opentrons.protocols.execution.types import LoadedLabware, Instruments
from opentrons_shared_data.protocol.constants import (
    JsonRobotCommand,
    JsonPipetteCommand,
)


if TYPE_CHECKING:
    from opentrons_shared_data.protocol.types import (
        JsonProtocolV4,
        JsonProtocolV5,
        MagneticModuleEngageParams,
        ModuleIDParams,
        TemperatureParams,
        ThermocyclerSetTargetBlockParams,
        ThermocyclerRunProfileParams,
        Command,
        TemperatureModuleCommandId,
        MagneticModuleCommandId,
        ThermocyclerCommandId,
    )
    from opentrons.protocols.execution.dev_types import (
        PipetteDispatch,
        JsonV4MagneticModuleDispatch,
        JsonV4TemperatureModuleDispatch,
        JsonV4ThermocyclerDispatch,
    )

MODULE_LOG = logging.getLogger(__name__)

# Special string for slot used by thermocycler in JSON protocols.
TC_SPANNING_SLOT = "span7_8_10_11"


def load_labware_from_json_defs(
    ctx: ProtocolContext, protocol: "JsonProtocolV4", modules: Dict[str, ModuleContext]
) -> LoadedLabware:
    protocol_labware = protocol["labware"]
    definitions = protocol["labwareDefinitions"]
    loaded_labware = {}

    for labware_id, props in protocol_labware.items():
        slot = props["slot"]
        definition = definitions[props["definitionId"]]
        label = props.get("displayName", None)
        if slot in modules:
            loaded_labware[labware_id] = modules[slot].load_labware_from_definition(
                definition, label
            )
        else:
            loaded_labware[labware_id] = ctx.load_labware_from_definition(
                definition, slot, label
            )

    return loaded_labware


def load_modules_from_json(
    ctx: ProtocolContext, protocol: "JsonProtocolV4"
) -> Dict[str, ModuleContext]:
    module_data = protocol["modules"]
    modules_by_id: Dict[str, ModuleContext] = {}

    # the sort order doesn't matter, we just need it to be stable
    # to ensure `load_module` side-effects are deterministic
    # (eg, multiple module support)
    sorted_modules = sorted(
        module_data.items(), key=lambda v: v[1]["slot"] + v[1]["model"]
    )

    for module_id, module_props in sorted_modules:
        model = module_props["model"]
        slot = module_props["slot"]
        if slot == TC_SPANNING_SLOT:
            instr = ctx.load_module(model)
        else:
            instr = ctx.load_module(model, slot)
        modules_by_id[module_id] = instr

    return modules_by_id


def _engage_magnet(
    module: MagneticModuleContext, params: "MagneticModuleEngageParams"
) -> None:
    engage_height = params["engageHeight"]
    module.engage(height_from_base=engage_height)


def _disengage_magnet(module: MagneticModuleContext, params: "ModuleIDParams") -> None:
    module.disengage()


def _temperature_module_set_temp(
    module: TemperatureModuleContext, params: "TemperatureParams"
) -> None:
    temperature = params["temperature"]
    module.start_set_temperature(temperature)


def _temperature_module_deactivate(
    module: TemperatureModuleContext, params: "ModuleIDParams"
) -> None:
    module.deactivate()


def _temperature_module_await_temp(
    module: TemperatureModuleContext, params: "TemperatureParams"
) -> None:
    temperature = params["temperature"]
    module.await_temperature(temperature)


def _thermocycler_close_lid(
    module: ThermocyclerContext, params: "ModuleIDParams"
) -> None:
    module.close_lid()


def _thermocycler_open_lid(
    module: ThermocyclerContext, params: "ModuleIDParams"
) -> None:
    module.open_lid()


def _thermocycler_deactivate_block(
    module: ThermocyclerContext, params: "ModuleIDParams"
) -> None:
    module.deactivate_block()


def _thermocycler_deactivate_lid(
    module: ThermocyclerContext, params: "ModuleIDParams"
) -> None:
    module.deactivate_lid()


def _thermocycler_set_block_temperature(
    module: ThermocyclerContext, params: "ThermocyclerSetTargetBlockParams"
) -> None:
    temperature = params["temperature"]
    module.set_block_temperature(temperature)


def _thermocycler_set_lid_temperature(
    module: ThermocyclerContext, params: "TemperatureParams"
) -> None:
    temperature = params["temperature"]
    module.set_lid_temperature(temperature)


def _thermocycler_run_profile(
    module: ThermocyclerContext, params: "ThermocyclerRunProfileParams"
) -> None:
    volume = params["volume"]
    profile = [
        ThermocyclerStep(temperature=p["temperature"], hold_time_seconds=p["holdTime"])
        for p in params["profile"]
    ]
    module.execute_profile(steps=profile, block_max_volume=volume, repetitions=1)


def assert_no_async_tc_behavior(commands: List["Command"]) -> None:
    # awaiters[i] corresponds to setters[i]
    setters = [
        "thermocycler/setTargetBlockTemperature",
        "thermocycler/setTargetLidTemperature",
        "thermocycler/runProfile",
    ]
    awaiters = [
        "thermocycler/awaitBlockTemperature",
        "thermocycler/awaitLidTemperature",
        "thermocycler/awaitProfileComplete",
    ]

    command_types = [c["command"] for c in commands]

    first_command = command_types[0]
    if first_command in awaiters:
        raise RuntimeError(
            (
                "Unsupported behavior. Cannot {} as the "
                + "first command of a protocol"
            ).format(first_command)
        )

    last_command = command_types[-1]
    if last_command in setters:
        raise RuntimeError(
            (
                "Unsupported behavior. Cannot {} as the " + "last command of a protocol"
            ).format(last_command)
        )

    # [a, b, c, d] -> [a, b], [b, c], [c, d]
    for a, b in zip(command_types, command_types[1:]):
        #  a setter must be immediately followed
        # by its corresponding awaiter
        if a in setters:
            expected_awaiter = awaiters[setters.index(a)]
            if b != expected_awaiter:
                raise RuntimeError(
                    (
                        "Unsupported behavior. {} must be immediately"
                        + "followed by {}"
                    ).format(a, expected_awaiter)
                )
        # an awaiter must be preceded by a setter
        elif b in awaiters:
            expected_setter = setters[awaiters.index(b)]
            raise RuntimeError(
                (
                    "Unsupported behavior. {} must be immediately " + "preceded by {}"
                ).format(b, expected_setter)
            )


def assert_tc_commands_do_not_use_unimplemented_params(
    commands: List["Command"],
) -> None:
    # raise errors if commands include optional param keys that
    # the executor/api does not currently support
    unsupported_keys_by_command = {"thermocycler/setTargetBlockTemperature": ["volume"]}
    for c in commands:
        command_type = c["command"]
        params = c["params"]
        unsupported_keys = unsupported_keys_by_command.get(command_type, [])
        for k in unsupported_keys:
            if k in params:
                raise RuntimeError(
                    (
                        "{} does not support {} param. "
                        + "This may be implemented in a later version of "
                        + "the robot server"
                    ).format(command_type, k)
                )


def dispatch_json(
    context: ProtocolContext,
    protocol_data: Union["JsonProtocolV4", "JsonProtocolV5"],
    instruments: Instruments,
    loaded_labware: LoadedLabware,
    modules: Dict[str, ModuleContext],
    pipette_command_map: "PipetteDispatch",
    magnetic_module_command_map: "JsonV4MagneticModuleDispatch",
    temperature_module_command_map: "JsonV4TemperatureModuleDispatch",
    thermocycler_module_command_map: "JsonV4ThermocyclerDispatch",
) -> None:
    commands = protocol_data["commands"]

    assert_no_async_tc_behavior(commands)
    assert_tc_commands_do_not_use_unimplemented_params(commands)

    for command_item in commands:
        command_type = command_item["command"]
        params = command_item["params"]
        # because of https://github.com/python/mypy/issues/8940
        # we can't narrow down types using in sadly
        if command_type in pipette_command_map:
            pipette_command_map[command_type](  # type: ignore
                instruments, loaded_labware, params
            )
        elif command_type in magnetic_module_command_map:
            handleMagnetCommand(
                params,  # type: ignore
                modules,
                command_type,  # type: ignore
                magnetic_module_command_map,  # type: ignore[arg-type]
            )
        elif command_type in temperature_module_command_map:
            handleTemperatureCommand(
                params,  # type: ignore[arg-type]
                modules,
                command_type,  # type: ignore
                temperature_module_command_map,  # type: ignore[arg-type]
            )
        elif command_type in thermocycler_module_command_map:
            handleThermocyclerCommand(
                params,  # type: ignore
                modules,  # type: ignore
                command_type,  # type: ignore
                thermocycler_module_command_map,  # type: ignore[arg-type]
            )
        elif command_item["command"] == JsonRobotCommand.delay.value:
            _delay(context, params)  # type: ignore
        elif command_type == JsonPipetteCommand.moveToSlot.value:
            _move_to_slot(context, instruments, params)  # type: ignore
        else:
            raise RuntimeError("Unsupported command type {}".format(command_type))


def handleTemperatureCommand(
    params: Union["TemperatureParams", "ModuleIDParams"],
    modules: Dict[str, ModuleContext],
    command_type: "TemperatureModuleCommandId",
    temperature_module_command_map: TemperatureModuleCommand,
) -> None:
    module_id = params["module"]
    module = modules[module_id]
    if isinstance(module, TemperatureModuleContext):
        temperature_module_command_map[command_type](module, params)  # type: ignore[typeddict-item]
    else:
        raise RuntimeError(
            "Temperature Module does not match " + "TemperatureModuleContext interface"
        )


def handleThermocyclerCommand(
    params: Union[
        "ModuleIDParams",
        "TemperatureParams",
        "ThermocyclerRunProfileParams",
        "ThermocyclerSetTargetBlockParams",
    ],
    modules: Dict[str, ThermocyclerContext],
    command_type: "ThermocyclerCommandId",
    thermocycler_module_command_map: ThermocyclerCommand,
) -> None:
    module_id = params["module"]
    module = modules[module_id]
    if isinstance(module, ThermocyclerContext):
        thermocycler_module_command_map[command_type](module, params)  # type: ignore[typeddict-item]
    else:
        raise RuntimeError(
            "Thermocycler Module does not match ThermocyclerContext interface"
        )


def handleMagnetCommand(
    params: Union["ModuleIDParams", "MagneticModuleEngageParams"],
    modules: Dict[str, ModuleContext],
    command_type: "MagneticModuleCommandId",
    magnetic_module_command_map: MagneticModuleCommand,
) -> None:
    module_id = params["module"]
    module = modules[module_id]
    if isinstance(module, MagneticModuleContext):
        magnetic_module_command_map[command_type](module, params)  # type: ignore[typeddict-item]
    else:
        raise RuntimeError(
            "Magnetic Module does not match MagneticModuleContext interface"
        )



"""Post-protocol hardware stopper."""
import logging
from typing import Optional

from opentrons.hardware_control import HardwareControlAPI
from opentrons.types import PipetteNotAttachedError as HwPipetteNotAttachedError

from ..resources.ot3_validation import ensure_ot3_hardware
from ..state.state import StateStore
from ..types import MotorAxis, PostRunHardwareState
from ..errors import HardwareNotSupportedError

from .movement import MovementHandler
from .gantry_mover import HardwareGantryMover
from .tip_handler import TipHandler, HardwareTipHandler
from ...hardware_control.types import OT3Mount

from opentrons.protocol_engine.types import AddressableOffsetVector

log = logging.getLogger(__name__)

# TODO(mc, 2022-03-07): this constant dup'd from opentrons.protocols.geometry.deck
# to avoid a circular dependency that needs to be figured out.
FIXED_TRASH_ID = "fixedTrash"


class HardwareStopper:
    """Class to implement hardware stopping."""

    def __init__(
        self,
        hardware_api: HardwareControlAPI,
        state_store: StateStore,
        movement: Optional[MovementHandler] = None,
        tip_handler: Optional[TipHandler] = None,
    ) -> None:
        """Hardware stopper initializer."""
        self._hardware_api = hardware_api
        self._state_store = state_store
        self._movement_handler = movement or MovementHandler(
            hardware_api=hardware_api,
            state_store=state_store,
            gantry_mover=HardwareGantryMover(
                hardware_api=hardware_api,
                state_view=state_store,
            ),
        )
        self._tip_handler = tip_handler or HardwareTipHandler(
            hardware_api=hardware_api,
            state_view=state_store,
        )

    async def _home_everything_except_plungers(self) -> None:
        # TODO: Update this once gripper MotorAxis is available in engine.
        try:
            ot3api = ensure_ot3_hardware(hardware_api=self._hardware_api)
            if (
                not self._state_store.config.use_virtual_gripper
                and ot3api.has_gripper()
            ):
                await ot3api.home_z(mount=OT3Mount.GRIPPER)
        except HardwareNotSupportedError:
            pass
        await self._movement_handler.home(
            axes=[MotorAxis.X, MotorAxis.Y, MotorAxis.LEFT_Z, MotorAxis.RIGHT_Z]
        )

    async def _try_to_drop_tips(self) -> None:
        """Drop currently attached tip, if any, into trash after a run cancel."""
        attached_tips = self._state_store.pipettes.get_all_attached_tips()

        if attached_tips:
            await self._hardware_api.stop(home_after=False)

            await self._home_everything_except_plungers()

            for pipette_id, tip in attached_tips:
                try:
                    if self._state_store.labware.get_fixed_trash_id() == FIXED_TRASH_ID:
                        # OT-2 and Flex 2.15 protocols will default to the Fixed Trash Labware
                        self._tip_handler.cache_tip(pipette_id=pipette_id, tip=tip)
                        await self._movement_handler.move_to_well(
                            pipette_id=pipette_id,
                            labware_id=FIXED_TRASH_ID,
                            well_name="A1",
                        )
                        await self._tip_handler.drop_tip(
                            pipette_id=pipette_id,
                            home_after=False,
                        )
                    elif self._state_store.config.robot_type == "OT-2 Standard":
                        # API 2.16 and above OT2 protocols use addressable areas
                        self._tip_handler.cache_tip(pipette_id=pipette_id, tip=tip)
                        await self._movement_handler.move_to_addressable_area(
                            pipette_id=pipette_id,
                            addressable_area_name="fixedTrash",
                            offset=AddressableOffsetVector(x=0, y=0, z=0),
                            force_direct=False,
                            speed=None,
                            minimum_z_height=None,
                        )
                        await self._tip_handler.drop_tip(
                            pipette_id=pipette_id,
                            home_after=False,
                        )
                    else:
                        log.debug(
                            "Flex Protocols API Version 2.16 and beyond do not support automatic tip dropping at this time."
                        )

                except HwPipetteNotAttachedError:
                    # this will happen normally during protocol analysis, but
                    # should not happen during an actual run
                    log.debug(f"Pipette ID {pipette_id} no longer attached.")

    async def do_halt(self, disengage_before_stopping: bool = False) -> None:
        """Issue a halt signal to the hardware API.

        After issuing a halt, you must call do_stop_and_recover after
        anything using the HardwareAPI has settled.
        """
        await self._hardware_api.halt(
            disengage_before_stopping=disengage_before_stopping
        )

    async def do_stop_and_recover(
        self,
        post_run_hardware_state: PostRunHardwareState,
        drop_tips_after_run: bool = False,
    ) -> None:
        """Stop and reset the HardwareAPI, homing and dropping tips independently if specified."""
        home_after_stop = post_run_hardware_state in (
            PostRunHardwareState.HOME_AND_STAY_ENGAGED,
            PostRunHardwareState.HOME_THEN_DISENGAGE,
        )
        if drop_tips_after_run:
            await self._try_to_drop_tips()

        await self._hardware_api.stop(home_after=False)

        if home_after_stop:
            await self._home_everything_except_plungers()



"""Opentrons analyze CLI."""
import click

from anyio import run
from contextlib import contextmanager
from dataclasses import dataclass
from datetime import datetime, timezone
from enum import Enum
from pathlib import Path
from pydantic import BaseModel
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Sequence,
    Union,
    Literal,
    Callable,
    IO,
    TypeVar,
    Iterator,
)
import logging
import sys
import json

from opentrons.protocol_engine.types import (
    RunTimeParameter,
    CSVRuntimeParamPaths,
    PrimitiveRunTimeParamValuesType,
    EngineStatus,
)
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocol_reader import (
    ProtocolReader,
    ProtocolFileRole,
    ProtocolType,
    JsonProtocolConfig,
    ProtocolFilesInvalidError,
    ProtocolSource,
)
from opentrons.protocol_runner.create_simulating_orchestrator import (
    create_simulating_orchestrator,
)
from opentrons.protocol_runner import RunResult
from opentrons.protocol_runner.run_orchestrator import ParseMode

from opentrons.protocol_engine import (
    Command,
    ErrorOccurrence,
    LoadedLabware,
    LoadedPipette,
    LoadedModule,
    Liquid,
    LiquidClassRecordWithId,
    StateSummary,
)
from opentrons.protocol_engine.protocol_engine import code_in_error_tree
from opentrons.protocol_engine.types import CommandAnnotation

from opentrons_shared_data.robot.types import RobotType

from opentrons_shared_data.errors import ErrorCodes
from opentrons_shared_data.errors.exceptions import (
    EnumeratedError,
    PythonException,
)


OutputKind = Literal["json", "human-json"]


@dataclass(frozen=True)
class _Output:
    to_file: IO[bytes]
    kind: OutputKind


@click.command()
@click.argument(
    "files",
    nargs=-1,
    required=True,
    type=click.Path(exists=True, path_type=Path, file_okay=True, dir_okay=True),
)
@click.option(
    "--json-output",
    help="Return analysis results as machine-readable JSON. Specify --json-output=- to use stdout, but be aware that Python protocols may contain print() which will make the output JSON invalid.",
    type=click.File(mode="wb"),
)
@click.option(
    "--human-json-output",
    help="Return analysis results as JSON, formatted for human eyes. Specify --human-json-output=- to use stdout, but be aware that Python protocols may contain print() which will make the output JSON invalid.",
    type=click.File(mode="wb"),
)
@click.option(
    "--check",
    help="Fail (via exit code) if the protocol had an error. If not specified, always succeed.",
    is_flag=True,
    default=False,
)
@click.option(
    "--log-output",
    help="Where to send logs. Can be a path, - for stdout, or stderr for stderr.",
    default="stderr",
    type=str,
)
@click.option(
    "--log-level",
    help="Level of logs to capture.",
    type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR"], case_sensitive=False),
    default="WARNING",
)
@click.option(
    "--rtp-values",
    help="Serialized JSON of runtime parameter variable names to values.",
    default="{}",
    type=str,
)
@click.option(
    "--rtp-files",
    help="Serialized JSON of runtime parameter variable names to file paths.",
    default="{}",
    type=str,
)
def analyze(
    files: Sequence[Path],
    rtp_values: str,
    rtp_files: str,
    json_output: Optional[IO[bytes]],
    human_json_output: Optional[IO[bytes]],
    log_output: str,
    log_level: str,
    check: bool,
) -> int:
    """Analyze a protocol.

    You can use `opentrons analyze` to get a protocol's expected
    equipment and commands.
    """
    outputs = _get_outputs(json=json_output, human_json=human_json_output)
    if not outputs and not check:
        raise click.UsageError(
            message="Please specify at least --check or one of the output options."
        )

    try:
        with _capture_logs(log_output, log_level):
            sys.exit(run(_analyze, files, rtp_values, rtp_files, outputs, check))
    except click.ClickException:
        raise
    except Exception as e:
        raise click.ClickException(str(e))


@contextmanager
def _capture_logs_to_stream(stream: IO[str]) -> Iterator[None]:
    handler = logging.StreamHandler(stream)
    logging.getLogger().addHandler(handler)
    try:
        yield
    finally:
        logging.getLogger().removeHandler(handler)


@contextmanager
def _capture_logs_to_file(filepath: Path) -> Iterator[None]:
    handler = logging.FileHandler(filepath, mode="w")
    logging.getLogger().addHandler(handler)
    try:
        yield
    finally:
        logging.getLogger().removeHandler(handler)


@contextmanager
def _capture_logs(write_to: str, log_level: str) -> Iterator[None]:
    try:
        level = getattr(logging, log_level)
    except AttributeError:
        raise click.ClickException(f"No such log level {log_level}")
    logging.getLogger().setLevel(level)
    if write_to in ("-", "stdout"):
        with _capture_logs_to_stream(sys.stdout):
            yield
    elif write_to == "stderr":
        with _capture_logs_to_stream(sys.stderr):
            yield
    else:
        with _capture_logs_to_file(Path(write_to)):
            yield


def _get_outputs(
    json: Optional[IO[bytes]],
    human_json: Optional[IO[bytes]],
) -> List[_Output]:
    outputs: List[_Output] = []
    if json:
        outputs.append(_Output(to_file=json, kind="json"))
    if human_json:
        outputs.append(_Output(to_file=human_json, kind="human-json"))
    return outputs


def _get_input_files(files_and_dirs: Sequence[Path]) -> List[Path]:
    results: List[Path] = []

    for entry in files_and_dirs:
        if entry.is_dir():
            results.extend(entry.glob("**/*"))
        else:
            results.append(entry)

    return results


def _get_runtime_parameter_values(
    serialized_rtp_values: str,
) -> PrimitiveRunTimeParamValuesType:
    rtp_values = {}
    try:
        for variable_name, value in json.loads(serialized_rtp_values).items():
            if not isinstance(value, (bool, int, float, str)):
                raise click.BadParameter(
                    f"Runtime parameter '{value}' is not of allowed type boolean, integer, float or string",
                    param_hint="--rtp-values",
                )
            rtp_values[variable_name] = value
    except json.JSONDecodeError as error:
        raise click.BadParameter(
            f"JSON decode error: {error}", param_hint="--rtp-values"
        )
    return rtp_values


def _get_runtime_parameter_paths(serialized_rtp_files: str) -> CSVRuntimeParamPaths:
    try:
        return {
            variable_name: Path(path_string)
            for variable_name, path_string in json.loads(serialized_rtp_files).items()
        }
    except json.JSONDecodeError as error:
        raise click.BadParameter(
            f"JSON decode error: {error}", param_hint="--rtp-files"
        )


R = TypeVar("R")


def _call_for_output_of_kind(
    kind: OutputKind, outputs: Sequence[_Output], fn: Callable[[IO[bytes]], R]
) -> Optional[R]:
    for output in outputs:
        if output.kind == kind:
            return fn(output.to_file)
    return None


def _get_return_code(analysis: RunResult) -> int:
    if analysis.state_summary.errors:
        return -1
    return 0


class UnexpectedAnalysisError(EnumeratedError):
    """An error raised while setting up the runner for analysis."""

    def __init__(
        self,
        message: Optional[str] = None,
        wrapping: Optional[Sequence[Union[EnumeratedError, Exception]]] = None,
    ) -> None:
        """Build a UnexpectedAnalysisError exception."""

        def _convert_exc() -> Iterator[EnumeratedError]:
            if not wrapping:
                return
            for exc in wrapping:
                if isinstance(exc, EnumeratedError):
                    yield exc
                else:
                    yield PythonException(exc)

        super().__init__(
            code=ErrorCodes.GENERAL_ERROR,
            message=message,
            wrapping=[e for e in _convert_exc()],
        )


async def _do_analyze(
    protocol_source: ProtocolSource,
    rtp_values: PrimitiveRunTimeParamValuesType,
    rtp_paths: CSVRuntimeParamPaths,
) -> RunResult:

    orchestrator = await create_simulating_orchestrator(
        robot_type=protocol_source.robot_type, protocol_config=protocol_source.config
    )
    try:
        await orchestrator.load(
            protocol_source=protocol_source,
            parse_mode=ParseMode.NORMAL,
            run_time_param_values=rtp_values,
            run_time_param_paths=rtp_paths,
        )
    except Exception as error:
        err_id = "analysis-setup-error"
        err_created_at = datetime.now(tz=timezone.utc)
        if isinstance(error, EnumeratedError):
            error_occ = ErrorOccurrence.from_failed(
                id=err_id, createdAt=err_created_at, error=error
            )
        else:
            enumerated_wrapper = UnexpectedAnalysisError(
                message=str(error),
                wrapping=[error],
            )
            error_occ = ErrorOccurrence.from_failed(
                id=err_id, createdAt=err_created_at, error=enumerated_wrapper
            )
        analysis = RunResult(
            commands=[],
            state_summary=StateSummary(
                errors=[error_occ],
                status=EngineStatus.IDLE,
                labware=[],
                pipettes=[],
                modules=[],
                labwareOffsets=[],
                liquids=[],
                wells=[],
                hasEverEnteredErrorRecovery=False,
                files=[],
                liquidClasses=[],
            ),
            parameters=[],
            command_annotations=[],
        )
        return analysis
    return await orchestrator.run(deck_configuration=[])


async def _analyze(
    files_and_dirs: Sequence[Path],
    rtp_values: str,
    rtp_files: str,
    outputs: Sequence[_Output],
    check: bool,
) -> int:
    input_files = _get_input_files(files_and_dirs)
    parsed_rtp_values = _get_runtime_parameter_values(rtp_values)
    rtp_paths = _get_runtime_parameter_paths(rtp_files)

    try:
        protocol_source = await ProtocolReader().read_saved(
            files=input_files,
            directory=None,
        )
    except ProtocolFilesInvalidError as error:
        raise click.ClickException(str(error))

    analysis = await _do_analyze(protocol_source, parsed_rtp_values, rtp_paths)
    return_code = _get_return_code(analysis)

    if not outputs:
        return return_code

    if len(analysis.state_summary.errors) > 0:
        if any(
            code_in_error_tree(
                root_error=error, code=ErrorCodes.RUNTIME_PARAMETER_VALUE_REQUIRED
            )
            for error in analysis.state_summary.errors
        ):
            result = AnalysisResult.PARAMETER_VALUE_REQUIRED
        else:
            result = AnalysisResult.NOT_OK
    else:
        result = AnalysisResult.OK

    results = AnalyzeResults.model_construct(
        createdAt=datetime.now(tz=timezone.utc),
        files=[
            ProtocolFile.model_construct(name=f.path.name, role=f.role)
            for f in protocol_source.files
        ],
        config=(
            JsonConfig.model_construct(
                schemaVersion=protocol_source.config.schema_version
            )
            if isinstance(protocol_source.config, JsonProtocolConfig)
            else PythonConfig.model_construct(
                apiVersion=protocol_source.config.api_version
            )
        ),
        result=result,
        metadata=protocol_source.metadata,
        robotType=protocol_source.robot_type,
        runTimeParameters=analysis.parameters,
        commands=analysis.commands,
        errors=analysis.state_summary.errors,
        labware=analysis.state_summary.labware,
        pipettes=analysis.state_summary.pipettes,
        modules=analysis.state_summary.modules,
        liquids=analysis.state_summary.liquids,
        commandAnnotations=analysis.command_annotations,
        liquidClasses=analysis.state_summary.liquidClasses,
    )

    _call_for_output_of_kind(
        "json",
        outputs,
        lambda to_file: to_file.write(
            results.model_dump_json(exclude_none=True).encode("utf-8"),
        ),
    )
    _call_for_output_of_kind(
        "human-json",
        outputs,
        lambda to_file: to_file.write(
            results.model_dump_json(exclude_none=True, indent=2).encode("utf-8")
        ),
    )
    if check:
        return return_code
    else:
        return 0


class ProtocolFile(BaseModel):
    """A file in a protocol analysis."""

    name: str
    role: ProtocolFileRole


class JsonConfig(BaseModel):
    """Configuration of a JSON protocol."""

    protocolType: Literal[ProtocolType.JSON] = ProtocolType.JSON
    schemaVersion: int


class PythonConfig(BaseModel):
    """Configuration of a Python protocol."""

    protocolType: Literal[ProtocolType.PYTHON] = ProtocolType.PYTHON
    apiVersion: APIVersion


class AnalysisResult(str, Enum):
    """Result of a completed protocol analysis.

    The result indicates whether the protocol is expected to run successfully.

    Properties:
        OK: No problems were found during protocol analysis.
        NOT_OK: Problems were found during protocol analysis. Inspect
            `analysis.errors` for error occurrences.
        PARAMETER_VALUE_REQUIRED: A value is required to be set for a parameter
            in order for the protocol to be analyzed/run. The absence of this does not
            inherently mean there are no parameters, as there may be defaults for all
            or unset parameters are not referenced or handled via try/except clauses.
    """

    OK = "ok"
    NOT_OK = "not-ok"
    PARAMETER_VALUE_REQUIRED = "parameter-value-required"


class AnalyzeResults(BaseModel):
    """Results of a protocol analysis.

    See robot-server's analysis models for field documentation.
    """

    # We want to unify this local analysis model with the one that robot-server returns.
    # Until that happens, we need to keep these fields in sync manually.

    # Fields that are currently unique to this local analysis module, missing from robot-server:
    createdAt: datetime
    files: List[ProtocolFile]
    config: Union[JsonConfig, PythonConfig]
    metadata: Dict[str, Any]

    # Fields that should match robot-server:
    result: AnalysisResult
    robotType: RobotType
    runTimeParameters: List[RunTimeParameter]
    commands: List[Command]
    labware: List[LoadedLabware]
    pipettes: List[LoadedPipette]
    modules: List[LoadedModule]
    liquids: List[Liquid]
    liquidClasses: List[LiquidClassRecordWithId]
    errors: List[ErrorOccurrence]
    commandAnnotations: List[CommandAnnotation]
